=head1 NAME 

XML::Application::Config - sponsor your application with an XML based config file

=head1 SYNOPSIS

use strict;
use XML::Application::Config;

my $cfg = new("myApplication","~/.myApplication/cfg.xml");

my $login  = $cfg->find('/auth/login/@value');

my $passwd = $cfg->find('/auth/password/@value');

=head1 DESCRIPTION

This module allows for an easy to use XML based config file.
The XML syntax allows for an <include/> statement, so 
one may modularize the configuration, e.g. for excluding 
authentication information from publicly viewable parts.

The following methods are provided:

=over 4

=cut

package XML::Application::Config;

$XML::Application::Config::VERSION = "0.2.3";

use strict;
use XML::XPath;
#dev# use Data::Dumper;

sub XML_BASE () { "/applicationconfig/application" }

=item new()

	my $cfg = new("myApplication","~/.myApplication/cfg.xml");

	The contructor. Takes two arguments: the application name 
	and the config file name(s).

=cut

sub new ($$) {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = {};
	$self->{application_name} = $_[0];
	$self->{xmlfilename}      = $_[1];
	$self->{xpath}            = undef;
	$self->{context}          = undef;
	$self->{error}            = undef;

	bless ($self,$class);

	if ( $self->readXMLConfigFile eq "0E0" ) {
		print $self->{error},"\n";
		die $self->{error};
	}
	$self->setupContext;
	$self->processIncludes;

	return $self;
}

=item readXMLConfigFile()

	$cfg->readXMLConfigFile();

	Reads and parses an XML formattet config file.
	Normally, you don't have to bother with that one.

=cut

sub readXMLConfigFile () {
	my $self = shift;
	if ( ! -f $self->{xmlfilename} ) {
		$self->{error} = "There is no file named '$self->{xmlfilename}'.";
		return "0E0";
	}
	$self->{xpath} = XML::XPath->new(filename => $self->{xmlfilename});
	$self->{cfg_format_version} = ($self->{xpath}->find('/applicationconfig/@version')->get_nodelist)[0]->string_value;
	#dev# print "V: ",$self->{cfg_format_version},"\n";
}

=item setupContext()

	$cfg->setupContext();

	Searches for the application specific XML context.
	Normally, you don't have to bother with that one.

=cut

sub setupContext () {
	my $self = shift;

	my $results = $self->{xpath}->find( XML_BASE . q#[@name='# . $self->{application_name} . q#']# );
	if (!$results->isa('XML::XPath::NodeSet')) {
	    die "ERROR: Found '$results' on ",ref $results,"\n";
	}
	$self->{context} = ($results->get_nodelist)[0];
}

=item cfgFormatVersion()

	$cfg->cfgFormatVersion();

	Returns the version of the XML frame structure.

=cut

sub cfgFormatVersion () {
	my $self = shift;
	
	$self->{cfg_format_version};
}

=item context()

	$cfg->context();

	Returns the XML::XPath context.
	Normally, you don't have to bother with that one.

=cut

sub context () {
	my $self = shift;
	
	$self->{context};
}

=item find('/x/path')

	my $passwd = $cfg->find('/auth/password/@value');

	Returns the string value of the given xpath prefixed by "/applicationconfig/application".
	You may take advantage of the full XPath syntax including XPath functions as specified by W3C.

=cut

sub find ($) {
	my $self = shift;
	my $xpath = $_[0];

	my $res = $self->{xpath}->find($xpath, $self->{context});

	#dev# print "ref: ",ref $res,"\n";
	if ( ref $res eq "XML::XPath::Number" ) {
		#dev# print "XML::XPath::Number = ",$res->string_value,"\n";
		return $res->string_value;
	} elsif ( ref $res eq "XML::XPath::NodeSet" or "anything else" ) {
		my @Nodelist = ($self->{xpath}->find($xpath, $self->{context})->get_nodelist);
		#dev# print "NL: ",ref $nodelist," ",$nodelist,"\n";
		if ( scalar @Nodelist >= 1 ) {
			return $Nodelist[0]->string_value;
		} else {
			warn "Warning: XPath '".XML_BASE."/$xpath' does not exist";
			return undef;
		}
	}
}

=item processIncludes()

	$cfg->processIncludes();

	Processes the <include/> tags recursively.
	Normally, you don't have to bother with that one.

=cut

sub processIncludes () {
	my $self = shift;

	my @Nodelist = ($self->{xpath}->find("//include", $self->{context})->get_nodelist);
	for my $n ( @Nodelist ) {
		# build current XPath from ancester stack
		# type: XML::XPath::Node::Element
		my @Path;
		for my $node ( $n->find('ancestor-or-self::*')->get_nodelist ) {
			#dev# print "node: ",ref $node,"\n";
			#dev# print "node: ",$node->getName,"\n";
			push @Path,$node->getName;
		}
		my $path = join "/","",@Path;

		my $includefile = $n->find('@file');
		my $acuteness   = $n->find('@acuteness');
		#dev# print "include $includefile $acuteness $path\n";
		if ( -e $includefile ) {
			#dev# print "XPATH-1: ",Dumper($self->{xpath});
			# on migration to XML::LibXML see xml-combine for this one
			#
			# read include file
			my $xinc = new XML::Application::Config( $self->{application_name}, $includefile );

			# search parent node
			my $pnode = $n->find('..')->[0]; 
			#dev# print "RP: " ,ref $pnode,"\n";
			#dev# print "PNODE: ",$pnode->getName,"\n";

			# remove <include/>-Tag
			$pnode->removeChild($n),"\n";
			#dev# print "XPATH-R: ",Dumper($self->{xpath});
			#dev# print "RI: " ,ref $xinc->context,"\n";
			
			# current context of include is "/applicationconfig/application"
			# so enter this one and copy all elements to current context of main tree
			for my $i ( $xinc->context->find('./*')->get_nodelist ) {
				#dev# print "Ri: " ,ref $i,"\n";
				$pnode->appendChild($i),"\n";
				#dev# print "XPATH-a: ",Dumper($self->{xpath});
			}
			#dev# print "XPATH-A: ",Dumper($self->{xpath});
		} elsif ( $acuteness eq "mandatory" ) {
			die "can't open include '$includefile'";
		}
	}
}

=item compareVersion()

	$cfg->compareVersion("0.2.2");

	Compare given version to $VERSION;

=cut

sub compareVersion ($) {
	my $self = shift;

	my @V1 =  split (/\./,$self::VERSION);
	my @V2 =  split (/\./,$_[0]);
	my $l = 0;
	my @V = map { length($_) > $l and $l = length($_)} (@V1, @V2);
	my $v1 = join "",map { sprintf "%0${l}d",$_ } @V1;
	my $v2 = join "",map { sprintf "%0${l}d",$_ } @V2;

	return $v1 >= $v2;
}

1;

=back

=head1 CFG.XML EXAMPLES

=head2 simple:

F<~/myApplication/cfg.xml>

  <?xml version="1.0" encoding="utf-8"?>
  <!-- Stand: Fri Mar 16 13:36:58 CET 2012 -->
  <applicationconfig version="0.2">
          <application name="myApplication">
  		  <hostname value="myHost" />
          </application>
  </applicationconfig>

=head2 with includes:

F<~/myApplication/cfg.xml>

  <?xml version="1.0" encoding="utf-8"?>
  <!-- Stand: Fri Mar 16 13:36:58 CET 2012 -->
  <applicationconfig version="0.2">
          <application name="myApplication">
                  <hostname value="myHost" />
                  <include file="~/myApplication/options.xml" acuteness="may" />
                  <include file="~/myApplication/auth.xml" acuteness="mandatory" />
          </application>
  </applicationconfig>

F<~/myApplication/options.xml>

  <?xml version="1.0" encoding="utf-8"?>
  <!-- Stand: Fri Mar 16 13:36:58 CET 2012 -->
  <applicationconfig version="0.2">
          <application name="myApplication">
  				<options>
  					<option name="port" val="1234"/>
  					<option name="cream" val="yes"/>
  				</options>
          </application>
  </applicationconfig>

F<~/myApplication/auth.xml>

  <?xml version="1.0" encoding="utf-8"?>
  <!-- Stand: Fri Mar 16 13:36:58 CET 2012 -->
  <applicationconfig version="0.2">
          <application name="myApplication">
  				<auth>
  					<username val="johndoe"/>
  					<password val="secret"/>
  				</auth>
          </application>
  </applicationconfig>


=head1 SEE ALSO

	XML Path Language (XPath) Version 1.0 http://http://www.w3.org/TR/xpath/ 
